use crate::{
    data::varint::read_varint,
    ProtocolError,
};
use std::io::Read;
use uuid::Uuid;

/// Packet data reader trait
pub trait DataReader {
    /// Read bytes
    fn read_bytes(&mut self, size: usize) -> Result<Vec<u8>, ProtocolError>;

    /// Read byte
    fn read_byte(&mut self) -> Result<u8, ProtocolError> {
        Ok(self.read_bytes(1)?[0])
    }
    /// Read String
    fn read_string(&mut self) -> Result<String, ProtocolError> {
        let size = self.read_usize_varint()?;
        match String::from_utf8(self.read_bytes(size)?) {
            Ok(i) => Ok(i),
            Err(_) => Err(ProtocolError::StringParseError),
        }
    }
    /// Read Unsigned Short as u16
    fn read_unsigned_short(&mut self) -> Result<u16, ProtocolError> {
        self.read_bytes(2)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| u16::from_be_bytes(o))
    }
    /// Read Boolean
    fn read_boolean(&mut self) -> Result<bool, ProtocolError> {
        self.read_byte().map(|o| o == 0x01)
    }
    /// Read Short as i16
    fn read_short(&mut self) -> Result<i16, ProtocolError> {
        self.read_bytes(2)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| i16::from_be_bytes(o))
    }
    /// Read Long as i64
    fn read_long(&mut self) -> Result<i64, ProtocolError> {
        self.read_bytes(8)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| i64::from_be_bytes(o))
    }
    /// Read Float as f32
    fn read_float(&mut self) -> Result<f32, ProtocolError> {
        self.read_bytes(4)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| f32::from_be_bytes(o))
    }
    /// Read Double as f64
    fn read_double(&mut self) -> Result<f64, ProtocolError> {
        self.read_bytes(8)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| f64::from_be_bytes(o))
    }
    /// Read Int as i32
    fn read_int(&mut self) -> Result<i32, ProtocolError> {
        self.read_bytes(4)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| i32::from_be_bytes(o))
    }
    /// Read UUID
    fn read_uuid(&mut self) -> Result<Uuid, ProtocolError> {
        self.read_bytes(16)
            .and_then(|o| o.try_into().map_err(|_| ProtocolError::ReadError))
            .map(|o| Uuid::from_bytes(o))
    }

    /// Read VarInt as usize with size in bytes (varint, size)
    fn read_usize_varint_size(&mut self) -> Result<(usize, usize), ProtocolError> {
        read_varint!(usize, self)
    }
    /// Read VarInt as u8 with size in bytes (varint, size)
    fn read_u8_varint_size(&mut self) -> Result<(u8, usize), ProtocolError> {
        read_varint!(u8, self)
    }
    /// Read VarInt as u16 with size in bytes (varint, size)
    fn read_u16_varint_size(&mut self) -> Result<(u16, usize), ProtocolError> {
        read_varint!(u16, self)
    }
    /// Read VarInt as u32 with size in bytes (varint, size)
    fn read_u32_varint_size(&mut self) -> Result<(u32, usize), ProtocolError> {
        read_varint!(u32, self)
    }
    /// Read VarInt as u64 with size in bytes (varint, size)
    fn read_u64_varint_size(&mut self) -> Result<(u64, usize), ProtocolError> {
        read_varint!(u64, self)
    }
    /// Read VarInt as u128 with size in bytes (varint, size)
    fn read_u128_varint_size(&mut self) -> Result<(u128, usize), ProtocolError> {
        read_varint!(u128, self)
    }

    /// Read VarInt as usize
    fn read_usize_varint(&mut self) -> Result<usize, ProtocolError> {
        self.read_usize_varint_size().map(|o| o.0)
    }
    /// Read VarInt as u8
    fn read_u8_varint(&mut self) -> Result<u8, ProtocolError> {
        self.read_u8_varint_size().map(|o| o.0)
    }
    /// Read VarInt as u16
    fn read_u16_varint(&mut self) -> Result<u16, ProtocolError> {
        self.read_u16_varint_size().map(|o| o.0)
    }
    /// Read VarInt as u32
    fn read_u32_varint(&mut self) -> Result<u32, ProtocolError> {
        self.read_u32_varint_size().map(|o| o.0)
    }
    /// Read VarInt as u64
    fn read_u64_varint(&mut self) -> Result<u64, ProtocolError> {
        self.read_u64_varint_size().map(|o| o.0)
    }
    /// Read VarInt as u128
    fn read_u128_varint(&mut self) -> Result<u128, ProtocolError> {
        self.read_u128_varint_size().map(|o| o.0)
    }

    /// Read VarInt as usize with zigzag
    fn read_isize_varint(&mut self) -> Result<isize, ProtocolError> {
        self.read_usize_varint().map(|o| o as isize)
    }
    /// Read VarInt as u8 with zigzag
    fn read_i8_varint(&mut self) -> Result<i8, ProtocolError> {
        self.read_u8_varint().map(|o| o as i8)
    }
    /// Read VarInt as u16 with zigzag
    fn read_i16_varint(&mut self) -> Result<i16, ProtocolError> {
        self.read_u16_varint().map(|o| o as i16)
    }
    /// Read VarInt as u32 with zigzag
    fn read_i32_varint(&mut self) -> Result<i32, ProtocolError> {
        self.read_u32_varint().map(|o| o as i32)
    }
    /// Read VarInt as u64 with zigzag
    fn read_i64_varint(&mut self) -> Result<i64, ProtocolError> {
        self.read_u64_varint().map(|o| o as i64)
    }
    /// Read VarInt as u128 with zigzag
    fn read_i128_varint(&mut self) -> Result<i128, ProtocolError> {
        self.read_u128_varint().map(|o| o as i128)
    }

    /// Read VarInt as usize with zigzag with size in bytes (varint, size)
    fn read_isize_varint_size(&mut self) -> Result<(isize, usize), ProtocolError> {
        self.read_usize_varint_size().map(|o| (o.0 as isize, o.1))
    }
    /// Read VarInt as u8 with zigzag with size in bytes (varint, size)
    fn read_i8_varint_size(&mut self) -> Result<(i8, usize), ProtocolError> {
        self.read_u8_varint_size().map(|o| (o.0 as i8, o.1))
    }
    /// Read VarInt as u16 with zigzag with size in bytes (varint, size)
    fn read_i16_varint_size(&mut self) -> Result<(i16, usize), ProtocolError> {
        self.read_u16_varint_size().map(|o| (o.0 as i16, o.1))
    }
    /// Read VarInt as u32 with zigzag with size in bytes (varint, size)
    fn read_i32_varint_size(&mut self) -> Result<(i32, usize), ProtocolError> {
        self.read_u32_varint_size().map(|o| (o.0 as i32, o.1))
    }
    /// Read VarInt as u64 with zigzag with size in bytes (varint, size)
    fn read_i64_varint_size(&mut self) -> Result<(i64, usize), ProtocolError> {
        self.read_u64_varint_size().map(|o| (o.0 as i64, o.1))
    }
    /// Read VarInt as u128 with zigzag with size in bytes (varint, size)
    fn read_i128_varint_size(&mut self) -> Result<(i128, usize), ProtocolError> {
        self.read_u128_varint_size().map(|o| (o.0 as i128, o.1))
    }
}

impl<R: Read> DataReader for R {
    fn read_bytes(&mut self, size: usize) -> Result<Vec<u8>, ProtocolError> {
        let mut buf = vec![0; size];
        match self.read(&mut buf) {
            Ok(i) => if i == size {
                Ok(buf)
            } else if i == 0 {
                Err(ProtocolError::ConnectionClosedError)
            } else {
                buf.truncate(i);
                buf.append(&mut self.read_bytes(size-i)?);
                Ok(buf)
            },
            Err(_) => Err(ProtocolError::ReadError),
        }
    }
}
