use uuid::Uuid;

use super::*;
use std::{io::Cursor, net::TcpListener, thread};

#[test]
fn test_varint() -> Result<(), ProtocolError> {
    let mut src = Packet::empty(0x00);

    src.write_u8_varint(0)?;               // 0x00
    src.write_u8_varint(1)?;               // 0x01
    src.write_u8_varint(2)?;               // 0x02
    src.write_u8_varint(127)?;             // 0x7f
    src.write_u8_varint(128)?;             // 0x80 0x01
    src.write_u16_varint(255)?;            // 0xff 0x01
    src.write_u16_varint(25565)?;          // 0xdd 0xc7 0x01
    src.write_u32_varint(2097151)?;        // 0xff 0xff 0x7f
    src.write_u32_varint(2147483647)?;     // 0xff 0xff 0xff 0xff 0x07
    src.write_i8_varint(-1)?;              // 0xff 0xff 0xff 0xff 0x0f
    src.write_i32_varint(-2147483648)?;    // 0x80 0x80 0x80 0x80 0x08

    let mut packet = Packet::from_bytes(0x00, src.get_bytes());
    assert_eq!(packet.read_u8_varint()?, 0);
    assert_eq!(packet.read_u8_varint()?, 1);
    assert_eq!(packet.read_u8_varint()?, 2);
    assert_eq!(packet.read_u8_varint()?, 127);
    assert_eq!(packet.read_u8_varint()?, 128);
    assert_eq!(packet.read_u16_varint()?, 255);
    assert_eq!(packet.read_u16_varint()?, 25565);
    assert_eq!(packet.read_u32_varint()?, 2097151);
    assert_eq!(packet.read_u32_varint()?, 2147483647);
    assert_eq!(packet.read_i8_varint()?, -1);
    assert_eq!(packet.read_i32_varint()?, -2147483648);

    let mut packet = Packet::from_bytes(0x00, src.get_bytes());
    assert_eq!(packet.read_bytes(1)?, vec![0x00]);                         // 0
    assert_eq!(packet.read_bytes(1)?, vec![0x01]);                         // 1
    assert_eq!(packet.read_bytes(1)?, vec![0x02]);                         // 2
    assert_eq!(packet.read_bytes(1)?, vec![0x7f]);                         // 127
    assert_eq!(packet.read_bytes(2)?, vec![0x80, 0x01]);                   // 128
    assert_eq!(packet.read_bytes(2)?, vec![0xff, 0x01]);                   // 255
    assert_eq!(packet.read_bytes(3)?, vec![0xdd, 0xc7, 0x01]);             // 25565
    assert_eq!(packet.read_bytes(3)?, vec![0xff, 0xff, 0x7f]);             // 2097151
    assert_eq!(packet.read_bytes(5)?, vec![0xff, 0xff, 0xff, 0xff, 0x07]); // 2147483647
    assert_eq!(packet.read_bytes(5)?, vec![0xff, 0xff, 0xff, 0xff, 0x0f]); // -1
    assert_eq!(packet.read_bytes(5)?, vec![0x80, 0x80, 0x80, 0x80, 0x08]); // -2147483648

    Ok(())
}

#[test]
fn test_data_transfer() -> Result<(), ProtocolError> {

    thread::spawn(move || -> Result<(), ProtocolError> {
        let listener =
            TcpListener::bind("localhost:44447").or(Err(ProtocolError::StreamConnectError))?;

        for stream in listener.incoming() {
            let mut stream = MCConnTcp::new(stream.or(Err(ProtocolError::StreamConnectError))?);

            stream.set_compression(Some(5));

            let mut packet = stream.read_packet()?;

            stream.write_packet(&Packet::build(packet.id(), |pack| {
                pack.write_boolean(packet.read_boolean()?)?;
                pack.write_byte(packet.read_byte()?)?;
                pack.write_bytes(&packet.read_bytes(10)?)?;
                pack.write_double(packet.read_double()?)?;
                pack.write_float(packet.read_float()?)?;
                pack.write_i128_varint(packet.read_i128_varint()?)?;
                pack.write_u128_varint(packet.read_u128_varint()?)?;
                pack.write_int(packet.read_int()?)?;
                pack.write_long(packet.read_long()?)?;
                pack.write_short(packet.read_short()?)?;
                pack.write_uuid(&packet.read_uuid()?)?;
                pack.write_string(&packet.read_string()?)?;
                Ok(())
            })?)?;

            stream.set_compression(None);

            let mut packet = stream.read_packet()?;

            stream.write_packet(&Packet::build(packet.id(), |pack| {
                pack.write_boolean(packet.read_boolean()?)?;
                pack.write_byte(packet.read_byte()?)?;
                pack.write_bytes(&packet.read_bytes(10)?)?;
                pack.write_double(packet.read_double()?)?;
                pack.write_float(packet.read_float()?)?;
                pack.write_i128_varint(packet.read_i128_varint()?)?;
                pack.write_u128_varint(packet.read_u128_varint()?)?;
                pack.write_int(packet.read_int()?)?;
                pack.write_long(packet.read_long()?)?;
                pack.write_short(packet.read_short()?)?;
                pack.write_uuid(&packet.read_uuid()?)?;
                pack.write_string(&packet.read_string()?)?;
                Ok(())
            })?)?;
        }

        Ok(())
    });
    
    let conn = MCConnTcp::connect("localhost:44447");

    while let Err(_) = conn {}

    let mut conn = conn?;

    conn.set_compression(Some(5));

    conn.write_packet(&Packet::build(0xfe, |pack| {
        pack.write_boolean(true)?;
        pack.write_byte(0x12)?;
        pack.write_bytes(&vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99])?;
        pack.write_double(123456789.123456789f64)?;
        pack.write_float(789456.44422f32)?;
        pack.write_i128_varint(468927513325566)?;
        pack.write_u128_varint(99859652365236523)?;
        pack.write_int(77861346i32)?;
        pack.write_long(789465123545678946i64)?;
        pack.write_short(1233i16)?;
        pack.write_uuid(&Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?)?;
        pack.write_string("&packet.read_string()?")?;
        Ok(())
    })?)?;

    let mut packet = conn.read_packet()?;

    assert_eq!(packet.read_boolean()?, true);
    assert_eq!(packet.read_byte()?, 0x12);
    assert_eq!(packet.read_bytes(10)?, vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99]);
    assert_eq!(packet.read_double()?, 123456789.123456789f64);
    assert_eq!(packet.read_float()?, 789456.44422f32);
    assert_eq!(packet.read_i128_varint()?, 468927513325566);
    assert_eq!(packet.read_u128_varint()?, 99859652365236523);
    assert_eq!(packet.read_int()?, 77861346i32);
    assert_eq!(packet.read_long()?, 789465123545678946i64);
    assert_eq!(packet.read_short()?, 1233i16);
    assert_eq!(packet.read_uuid()?, Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?);
    assert_eq!(packet.read_string()?, "&packet.read_string()?");

    conn.set_compression(None);

    conn.write_packet(&Packet::build(0xfe, |pack| {
        pack.write_boolean(true)?;
        pack.write_byte(0x12)?;
        pack.write_bytes(&vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99])?;
        pack.write_double(123456789.123456789f64)?;
        pack.write_float(789456.44422f32)?;
        pack.write_i128_varint(468927513325566)?;
        pack.write_u128_varint(99859652365236523)?;
        pack.write_int(77861346i32)?;
        pack.write_long(789465123545678946i64)?;
        pack.write_short(1233i16)?;
        pack.write_uuid(&Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?)?;
        pack.write_string("&packet.read_string()?")?;
        Ok(())
    })?)?;

    let mut packet = conn.read_packet()?;

    assert_eq!(packet.read_boolean()?, true);
    assert_eq!(packet.read_byte()?, 0x12);
    assert_eq!(packet.read_bytes(10)?, vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99]);
    assert_eq!(packet.read_double()?, 123456789.123456789f64);
    assert_eq!(packet.read_float()?, 789456.44422f32);
    assert_eq!(packet.read_i128_varint()?, 468927513325566);
    assert_eq!(packet.read_u128_varint()?, 99859652365236523);
    assert_eq!(packet.read_int()?, 77861346i32);
    assert_eq!(packet.read_long()?, 789465123545678946i64);
    assert_eq!(packet.read_short()?, 1233i16);
    assert_eq!(packet.read_uuid()?, Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?);
    assert_eq!(packet.read_string()?, "&packet.read_string()?");

    Ok(())
}

#[test]
fn test_compression() -> Result<(), ProtocolError> {
    let mut conn = MCConn::new(Cursor::new(Vec::new()));
    conn.set_compression(Some(5));

    let mut packet_1 = Packet::empty(0x12);
    packet_1.write_bytes(b"1234567890qwertyuiopasdfghjklzxcvbnm")?;
    dbg!(&packet_1);
    conn.write_packet(&packet_1)?;

    conn.get_mut().set_position(0);

    let mut packet_2 = conn.read_packet()?;

    assert_eq!(
        packet_2.read_bytes(36)?,
        b"1234567890qwertyuiopasdfghjklzxcvbnm"
    );

    Ok(())
}
